from __future__ import annotations

from collections import defaultdict
from typing import TypeVar, Generic, Dict, Optional, Iterable, List, Iterator, Tuple, Generator, Union, Any, Set
from .parents import OuterProxy
from ..utils import object_trace
from ..utils import output, BColors

T = TypeVar('T', bound=OuterProxy)


class Collection(Generic[T]):
    __is_collection__ = True

    _data: List[T]

    _indexed_from_id: Dict[int, Dict[str, Any]]
    _indexed_values: Dict[str, Dict[Any, T]]

    shallow_list = property(fget=lambda self: self.data)

    def __init__(
            self,
            data: Optional[Iterable[T]] = None,
            sync_on_append: Dict[str, Collection] = None,
            append_object_to_attribute: Dict[str, T] = None,
            extend_object_to_attribute: Dict[str, Collection] = None,
    ) -> None:
        self._collection_for: dict = dict()

        self._contains_ids = set()
        self._data = []

        # List of collection attributes that should be modified on append
        # Key: collection attribute (str) of appended element
        # Value: main collection to sync to
        self.append_object_to_attribute: Dict[str, T] = append_object_to_attribute or {}
        self.extend_object_to_attribute: Dict[str, Collection[T]] = extend_object_to_attribute or {}
        self.sync_on_append: Dict[str, Collection] = sync_on_append or {}
        self.pull_from: List[Collection] = []
        self.push_to: List[Collection] = []

        self._id_to_index_values: Dict[int, set] = defaultdict(set)
        
        # This is to cleanly unmap previously mapped items by their id
        self._indexed_from_id: Dict[int, Dict[str, Any]] = defaultdict(dict)
        # this is to keep track and look up the actual objects
        self._indexed_values: Dict[str, Dict[Any, T]] = defaultdict(dict)

        self.extend(data)

    def __repr__(self) -> str:
        return f"Collection({id(self)})"

    def _map_element(self, __object: T, from_map: bool = False):
        self._unmap_element(__object.id)

        self._indexed_from_id[__object.id]["id"] = __object.id
        self._indexed_values["id"][__object.id] = __object

        for name, value in __object.indexing_values:
            if value is None or value == __object._inner._default_values.get(name):
                continue

            self._indexed_values[name][value] = __object
            self._indexed_from_id[__object.id][name] = value

    def _unmap_element(self, __object: Union[T, int]):
        obj_id = __object.id if isinstance(__object, OuterProxy) else __object

        if obj_id not in self._indexed_from_id:
            return

        for name, value in self._indexed_from_id[obj_id].items():
            if value in self._indexed_values[name]:
                del self._indexed_values[name][value]

        del self._indexed_from_id[obj_id]

    def _remap(self):
        for e in self:
            self._map_element(e)

    def _find_object(self, __object: T, no_push_to: bool = False) -> Optional[T]:
        if not no_push_to:
            for c in self.push_to:
                found, found_in = c._find_object(__object, no_push_to=True)
                if found is not None:
                    return found, found_in   

        self._remap()

        for name, value in __object.indexing_values:
            if value in self._indexed_values[name]:
                return self._indexed_values[name][value], self

        return None, self

    def append(self, __object: Optional[T], **kwargs):
        """
        If an object, that represents the same entity exists in a relevant collection,
        merge into this object. (and remap)
        Else append to this collection.

        :param __object:
        :return:
        """


        if __object is None:
            return

        existing_object, map_to = self._find_object(__object, no_push_to=kwargs.get("no_push_to", False))
        
        if map_to is self:
            for other, contained in (c._find_object(__object, no_push_to=True) for c in self.pull_from):
                output(other, __object, contained, color=BColors.RED)
                if other is None:
                    continue

                __object.__merge__(other, no_push_to=False, **kwargs)
                contained.remove(other)

        if existing_object is None:
            # append
            self._data.append(__object)
            self._map_element(__object)

            for collection_attribute, child_collection in self.extend_object_to_attribute.items():
                __object.__getattribute__(collection_attribute).extend(child_collection, **kwargs)

            for attribute, new_object in self.append_object_to_attribute.items():
                __object.__getattribute__(attribute).append(new_object, **kwargs)

            # only modify collections if the object actually has been appended
            for attribute, a in self.sync_on_append.items():
                b = __object.__getattribute__(attribute)
                if a is b:
                    continue

                no_sync_collection: Set[Collection] = kwargs.get("no_sync_collection", set())
                object_trace(f"Syncing [{a}] = [{b}]; {no_sync_collection}")
                if id(b) in no_sync_collection:
                    continue


                b_data = b.data.copy()
                b_collection_for = b._collection_for.copy()
                no_sync_collection.add(id(b))
                # kwargs["no_sync_collection"] = no_sync_collection
                del b

                a.extend(b_data, **kwargs)
                
                for synced_with, key in b_collection_for.items():
                    synced_with.__setattr__(key, a)
                    a._collection_for[synced_with] = key

        else:
            # merge only if the two objects are not the same
            if existing_object.id == __object.id:
                return

            existing_object.merge(__object, **kwargs)
            map_to._map_element(existing_object)  

    def remove(self, __object: T) -> T:
        self._data.remove(__object)
        self._unmap_element(__object)
        return __object

    def contains(self, __object: T) -> bool:
        return self._find_object(__object) is not None

    def extend(self, __iterable: Optional[Generator[T, None, None]], **kwargs):
        if __iterable is None:
            return

        for __object in __iterable:
            self.append(__object, **kwargs)

    @property
    def data(self) -> List[T]:
        return list(self.__iter__())

    def __len__(self) -> int:
        return len(self._data)

    @property
    def empty(self) -> bool:
        return self.__len__() <= 0

    def __iter__(self) -> Iterator[T]:
        yield from self._data

    def __merge__(self, __other: Collection, **kwargs):
        self.extend(__other, **kwargs)

    def __getitem__(self, item: int):
        return self._data[item]

    def get(self, item: int, default = None):
        if item >= len(self._data):
            return default
        return self._data[item]
